"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.bootstrap = void 0;
const joi_1 = __importDefault(require("joi"));
const container_1 = require("./container");
const enum_1 = require("./constants/enum");
const json_1 = require("./helpers/json");
const constants_1 = require("./constants/constants");
const authDecorator_1 = require("./decorators/authDecorator");
const authResponse_1 = require("./models/authResponse");
const request_1 = require("./models/request");
const endpoint_1 = require("./models/endpoint");
/**
 * Bootstraps the whole application.
 * @param params Params to assist with the bootstrapping process.
 */
function bootstrap(params) {
    params.controllers.forEach(x => {
        const controllerName = x.name.replace('Controller', '');
        const controllerInstance = new x();
        controllerInstance.name = controllerName;
        container_1.ControllersContainer.set(controllerName.toLowerCase(), controllerInstance);
    });
    const trimRegex = /(^\/)|(\/$)/g;
    params.base = params.base ? `/${params.base.replace(trimRegex, '')}` : '';
    return (req, res, next) => {
        let isRouteResolved = false;
        let validationErrorOccurred = false;
        const url = req.url.replace(params.base, '');
        const method = req.method;
        const controllerName = url.split('/')[1].split('?')[0];
        const queryParams = req.query;
        const requestBody = req.body;
        const controller = container_1.ControllersContainer.get(controllerName);
        if (controller) {
            const routes = controller.routes;
            const matchedRoute = routes.find(x => new RegExp(x.routeRegex).test(url)
                && x.httpMethod.toString() === method);
            if (matchedRoute && req.url.startsWith(params.base)) {
                // Initially, verify if the route has AuthNAuth enabled.
                // If so, call the appropriate AuthNAuth handler.
                let authResponse = new authResponse_1.AuthResponse(true, enum_1.HTTPResponse.Success);
                // Fetching auth info for all routes within the target controller.
                const authRoutes = controller[authDecorator_1.authRoutesPropertyName];
                // Fetching no auth info for all routes within the target controller.
                const noAuthRoutes = controller[authDecorator_1.noAuthRoutesPropertyName];
                // Fetching auth info for the controller itself.
                const controllerAuth = controller[authDecorator_1.controllerLevelAuthPropertyName];
                if (authRoutes) {
                    // Find the auth info for the target route/method.
                    const targetRouteAuthInfo = authRoutes.find(x => x.method === matchedRoute.method);
                    if (targetRouteAuthInfo) {
                        // If callback is assigned to the auth info, use it, otherwise use the one provided in this method if present.
                        const targetCallback = targetRouteAuthInfo.callback ? targetRouteAuthInfo.callback :
                            controllerAuth && controllerAuth.callback ? controllerAuth.callback : params.authCallback;
                        authResponse = determineAuthResponse(req, matchedRoute, controllerName, targetCallback);
                        // Since an @Auth decorator is not found on our target method, perhaps the whole controller is under Auth.
                    }
                    else if (controllerAuth) {
                        // The whole controller is under Auth. Now we need to check whether our method is excluded from Auth.
                        // Find the no auth info for the target route/method.
                        let targetRouteNoAuthInfo;
                        if (noAuthRoutes) {
                            targetRouteNoAuthInfo = noAuthRoutes.find(x => x.method === matchedRoute.method);
                        }
                        // Since our method is not found within NoAuth routes, we can proceed with the auth callback.
                        if (!targetRouteNoAuthInfo) {
                            const targetCallback = controllerAuth.callback ? controllerAuth.callback : params.authCallback;
                            authResponse = determineAuthResponse(req, matchedRoute, controllerName, targetCallback);
                        }
                    }
                    // Since an @Auth decorator is not found on any of the target controller's methods, perhaps the whole controller is under Auth.
                }
                else if (controllerAuth) {
                    // The whole controller is under Auth. Now we need to check whether our method is excluded from Auth.
                    // Find the no auth info for the target route/method.
                    let targetRouteNoAuthInfo;
                    if (noAuthRoutes) {
                        targetRouteNoAuthInfo = noAuthRoutes.find(x => x.method === matchedRoute.method);
                    }
                    // Since our method is not found within NoAuth routes, we can proceed with the auth callback.
                    if (!targetRouteNoAuthInfo) {
                        const targetCallback = controllerAuth.callback ? controllerAuth.callback : params.authCallback;
                        authResponse = determineAuthResponse(req, matchedRoute, controllerName, targetCallback);
                    }
                }
                if (!authResponse.isAuthenticatedAndAuthorized) {
                    return res.status(authResponse.statusCode).send(authResponse.authFailureMessage);
                }
                const routeParams = getRouteParams(matchedRoute.route, matchedRoute.routeParamsIndices, url);
                let methodExecutionExpression = `${Object.keys({ controller })[0]}.${matchedRoute.method}(`;
                let modelParams = new Array();
                const paramsArrayVarName = Object.keys({ modelParams })[0];
                matchedRoute.params.forEach((x, index) => {
                    let validationError;
                    switch (x.paramType) {
                        case enum_1.ParamType.Body:
                            validationError = performInputValidation(x.model, requestBody);
                            if (validationError) {
                                res.status(enum_1.HTTPResponse.UnprocessableEntity).send(validationError);
                                validationErrorOccurred = true;
                            }
                            else {
                                modelParams.push(json_1.deserialize(requestBody, x.model));
                                methodExecutionExpression += `${paramsArrayVarName}[${modelParams.length - 1}]`;
                            }
                            break;
                        case enum_1.ParamType.Query:
                            validationError = performInputValidation(x.model, queryParams);
                            if (validationError) {
                                res.status(enum_1.HTTPResponse.UnprocessableEntity).send(validationError);
                                validationErrorOccurred = true;
                            }
                            else {
                                modelParams.push(json_1.deserialize(queryParams, x.model));
                                methodExecutionExpression += `${paramsArrayVarName}[${modelParams.length - 1}]`;
                            }
                            break;
                        case enum_1.ParamType.Route:
                            methodExecutionExpression += isNaN(routeParams[x.name]) ?
                                `'${decodeURIComponent(routeParams[x.name])}'` : routeParams[x.name];
                            break;
                    }
                    if (index < matchedRoute.params.length - 1) {
                        methodExecutionExpression += ',';
                    }
                });
                methodExecutionExpression += ')';
                if (!validationErrorOccurred) {
                    const result = eval(methodExecutionExpression);
                    // Check if result has the custom 'Response' type.
                    if (result[constants_1.responseClassIdentifier]) {
                        res.status(result.statusCode).send(result.data);
                    }
                    else {
                        res.status(200).send(result);
                    }
                }
                isRouteResolved = true;
            }
        }
        if (!isRouteResolved) {
            res.status(404).send('Not Found');
        }
    };
}
exports.bootstrap = bootstrap;
function determineAuthResponse(req, matchedRoute, controllerName, targetCallback) {
    let authResponse = new authResponse_1.AuthResponse(true, enum_1.HTTPResponse.Success);
    if (targetCallback) {
        const request = new request_1.Request(req.headers, req.query, req.body);
        const endpoint = new endpoint_1.Endpoint(matchedRoute.route, controllerName, matchedRoute.method, matchedRoute.httpMethod);
        authResponse = targetCallback(request, endpoint);
    }
    return authResponse;
}
function getRouteParams(toMatch, routeParamsIndices, toParse) {
    const toMatchSplit = toMatch.split('/');
    const toParseSplit = toParse.split('/');
    const routeParams = new Object();
    routeParamsIndices.forEach(index => {
        const queryParamsRegex = /[?].*/;
        routeParams[toMatchSplit[index].replace(':', '')] =
            toParseSplit[index].replace(queryParamsRegex, '');
    });
    return routeParams;
}
function performInputValidation(model, input) {
    let validationErrorDetails;
    const obj = new model();
    if (obj[constants_1.validationSchemaPropertyName]) {
        const validationError = joi_1.default.object(obj[constants_1.validationSchemaPropertyName]).validate(input).error;
        if (validationError) {
            validationErrorDetails = validationError.details;
        }
    }
    return validationErrorDetails;
}
